import { Value, is, Position } from '@preserves/core';
import { ModulePath, Ref, Schema, $definitions, Definition, Alternative } from './gen/schema';
import { BASE } from './base';
import { SchemaSyntaxError } from './error';

export * from './gen/schema';

export type Input = Value<never>;

export function isValidToken(s: string, allowLeadingUnderscore = false): boolean {
    if (allowLeadingUnderscore) {
        return /^[a-zA-Z_][a-zA-Z_0-9]*$/.test(s);
    } else {
        return /^[a-zA-Z][a-zA-Z_0-9]*$/.test(s);
    }
}

export const ANDSYM = Symbol.for('&');
export const DOT = Symbol.for('.');
export const DOTDOTDOT = Symbol.for('...');
export const EQUALS = Symbol.for('=');
export const INCLUDE = Symbol.for('include');
export const ORSYM = Symbol.for('/');

export type SchemaEnvEntry = { schemaModulePath: ModulePath } & (
    ({
        typescriptModulePath: string | null, // null means it's "this module" in disguise
        schema: Schema,
    }) | ({
        typescriptModulePath: string,
        schema: null,
    })
);

export type Environment = Array<SchemaEnvEntry>;

function modsymFor(e: SchemaEnvEntry): string {
    return '_i_' + e.schemaModulePath.map(s => s.description!).join('$');
}

export function lookup<R>(namePos: Position | null,
                          name: Ref,
                          env: Environment,
                          kLocal: (p: Definition) => R,
                          kBase: (p: Alternative) => R,
                          kOther: (modId: string, modPath: string, p: Definition | null) => R): R
{
    for (const e of env) {
        if (is(e.schemaModulePath, Ref._.module(name)) ||
            (e.typescriptModulePath === null && Ref._.module(name).length === 0))
        {
            if (e.schema === null) {
                // It's an artificial module, not from a schema. Assume the identifier is present.
                return kOther(modsymFor(e), e.typescriptModulePath, null);
            } else {
                const p = Schema._._field0(e.schema).get($definitions).get(Ref._.name(name));
                if (p !== void 0) {
                    if (e.typescriptModulePath === null) {
                        return kLocal(p);
                    } else {
                        return kOther(modsymFor(e), e.typescriptModulePath, p);
                    }
                }
            }
        }
    }

    if (Ref._.module(name).length === 0) {
        const p = Schema._._field0(BASE).get($definitions).get(Ref._.name(name));
        if (p !== void 0) return kBase(p as Alternative);
    }

    throw new SchemaSyntaxError(`Undefined reference: ${formatRef(name)}`, namePos);
}

export function formatRef(r: Ref): string {
    return [... r[0], r[1]].map(s => s.description!).join('.');
}
